import os

from flask import (json, send_from_directory, current_app,
                   url_for as flask_url_for)

__all__ = ['Manifest']

CONFIG_PREFIX = 'MANIFEST_'
DEFAULT_CONFIG = {
    'MANIFEST_DEBUG': False,
    'MANIFEST_ROOT': 'dist',
    'MANIFEST_SERVE_ROOT': True,
    'MANIFEST_FILENAME': 'rev-manifest.json',
    'MANIFEST_ENDPOINT': 'manifest-dist',
    'MANIFEST_URL_PATH': '/dist',
    'MANIFEST_DOMAIN': None,
}
FOREVER = 10 * 365 * 24 * 60 * 60
EXTENSION_KEY = 'manifest'


def get_state():
    return current_app.extensions[EXTENSION_KEY]


def send_static_file(filename):
    return send_from_directory(get_state().root, filename,
                               cache_timeout=FOREVER)


def url_for(endpoint, **values):
    if endpoint == 'static':
        state = get_state()

        try:
            rev_file = state.manifest_contents[values['filename']]
        except KeyError:
            pass
        else:
            values['filename'] = rev_file

            if state.serve_root:
                endpoint = state.endpoint
            else:
                # TODO: Handle external domains
                values['_external'] = True

    return flask_url_for(endpoint, **values)


class ManifestState(object):
    def __init__(self, app):
        for key, value in DEFAULT_CONFIG.items():
            attr_name = key.replace(CONFIG_PREFIX, '').lower()
            setattr(self, attr_name, app.config.get(key, value))

        self.manifest = {}


class Manifest(object):
    def __init__(self, app=None):
        self.app = app

        if self.app is not None:
            self.init_app(app)

    def init_app(self, app):
        state = ManifestState(app)

        if app.debug and not state.debug:
            return

        app.extensions[EXTENSION_KEY] = state

        path = os.path.join(state.root, state.filename)

        with app.open_resource(path) as manifest:
            state.manifest_contents = json.load(manifest)

        if state.serve_root:
            app.add_url_rule(state.url_path + '/<path:filename>',
                             endpoint=state.endpoint,
                             view_func=send_static_file)

        app.jinja_env.globals['url_for'] = url_for



from cStringIO import StringIO
from contextlib import contextmanager
import mock
import pytest

from flask import Flask, current_app
from flask.ext.manifest import (Manifest, get_state, url_for, FOREVER,
                                EXTENSION_KEY)


TEST_MANIFEST = """{
  "image1.jpg": "image1-xyz.jpg",
  "image2.jpg": "image2-def.jpg"
}
"""


def create_test_app(debug=False, config=None):
    app = Flask(__name__)
    app.debug = debug

    if config:
        app.config.update(**config)

    Manifest(app)
    return app


@contextmanager
def open_manifest(*args):
    yield StringIO(TEST_MANIFEST)


@pytest.yield_fixture(autouse=True)
def app(monkeypatch):
    with mock.patch.object(Flask, 'open_resource', open_manifest):
        test_app = create_test_app()

    with test_app.test_request_context():
        yield test_app


def uses_app(func):
    return pytest.mark.usefixtures('app')


def test_manifest_contents_loaded():
    contents = get_state().manifest_contents
    assert 2 == len(contents)
    assert 'image1-xyz.jpg' == contents['image1.jpg']


def test_url_for():
    values = {'filename': 'image1.jpg'}
    assert '/dist/image1-xyz.jpg' == url_for('static', **values)


def test_url_for_fallback_to_flask_url_for():
    """
    If the ``filename`` isn't present in the manifest contents,
    the normal static URL should be returned.
    """
    assert '/static/image3.jpg' == url_for('static', filename='image3.jpg')


def test_manifest_endpoint_view_func():
    target = 'flask_manifest.send_from_directory'

    with current_app.test_client() as client:
        with mock.patch(target) as mock_send_file:
            mock_send_file.return_value = 'contents'
            response = client.get(url_for('static', filename='image1.jpg'))

    state = get_state()
    mock_send_file.assert_called_once_with(state.root, 'image1-xyz.jpg',
                                           cache_timeout=FOREVER)
    assert 'contents' == response.data


def test_extension_does_nothing_in_debug_mode(app):
    assert EXTENSION_KEY in app.extensions
    assert EXTENSION_KEY not in create_test_app(debug=True).extensions

    with mock.patch.object(Flask, 'open_resource', open_manifest):
        app = create_test_app(debug=True, config={'MANIFEST_DEBUG': True})

    assert EXTENSION_KEY in app.extensions


@pytest.mark.xfail
def test_url_for_without_serving_dist_root():
    """
    Some other server might be serving the files in the manifest.
    TODO: Figure out the best way to construct URLs with the
    ``MANIFEST_DOMAIN``.
    """
    state = get_state()
    state.serve_root = False
    state.domain = 'example.com'
    expected_url = 'http://example.com/dist/image1-xyz.jpg'
    assert expected_url == url_for('static', filename='image1.jpg')



"""
Flask-Manifest
-------------

Asset manifest integration with Flask.
"""
import codecs
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


def read(filename):
    return codecs.open(filename, 'r', encoding='utf-8').read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name='Flask-Manifest',
    version='0.2.0',
    url='https://github.com/ryankask/flask-manifest',
    license='BSD',
    author='Ryan Kaskel',
    author_email='dev@ryankaskel.com',
    description='Asset manifest integration with Flask.',
    long_description=read('README.rst'),
    py_modules=['flask_manifest'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=['Flask'],
    tests_require=['pytest', 'pytest-cov', 'mock'],
    cmdclass={
        'test': PyTest
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
